@@ -221,6 +221,11 @@ void dWorldCheck (dxWorld *w)
 
 //****************************************************************************
 // body
+dxWorld* dBodyGetWorld (dxBody* b)
+{
+  dAASSERT (b);
+  return b->world;
+}
 
 dxBody *dBodyCreate (dxWorld *w)
 {
@@ -230,16 +235,18 @@ dxBody *dBodyCreate (dxWorld *w)
   b->firstjoint = 0;
   b->flags = 0;
   b->geom = 0;
+  b->average_lvel_buffer = 0;
+  b->average_avel_buffer = 0;
   dMassSetParameters (&b->mass,1,0,0,0,1,1,1,0,0,0);
   dSetZero (b->invI,4*3);
   b->invI[0] = 1;
   b->invI[5] = 1;
   b->invI[10] = 1;
   b->invMass = 1;
-  dSetZero (b->pos,4);
+  dSetZero (b->posr.pos,4);
   dSetZero (b->q,4);
   b->q[0] = 1;
-  dRSetIdentity (b->R);
+  dRSetIdentity (b->posr.R);
   dSetZero (b->lvel,4);
   dSetZero (b->avel,4);
   dSetZero (b->facc,4);
@@ -249,9 +256,13 @@ dxBody *dBodyCreate (dxWorld *w)
   w->nb++;
 
   // set auto-disable parameters
+  b->average_avel_buffer = b->average_lvel_buffer = 0; // no buffer at beginnin
   dBodySetAutoDisableDefaults (b);	// must do this after adding to world
   b->adis_stepsleft = b->adis.idle_steps;
   b->adis_timeleft = b->adis.idle_time;
+  b->average_counter = 0;
+  b->average_ready = 0; // average buffer not filled on the beginning
+  dBodySetAutoDisableAverageSamplesCount(b, b->adis.average_samples);
 
   return b;
 }
@@ -284,6 +295,19 @@ void dBodyDestroy (dxBody *b)
   }
   removeObjectFromList (b);
   b->world->nb--;
+
+  // delete the average buffers
+  if(b->average_lvel_buffer)
+  {
+	  delete[] (b->average_lvel_buffer);
+	  b->average_lvel_buffer = 0;
+  }
+  if(b->average_avel_buffer)
+  {
+	  delete[] (b->average_avel_buffer);
+	  b->average_avel_buffer = 0;
+  }
+
   delete b;
 }
 
@@ -305,9 +329,9 @@ void *dBodyGetData (dBodyID b)
 void dBodySetPosition (dBodyID b, dReal x, dReal y, dReal z)
 {
   dAASSERT (b);
-  b->pos[0] = x;
-  b->pos[1] = y;
-  b->pos[2] = z;
+  b->posr.pos[0] = x;
+  b->posr.pos[1] = y;
+  b->posr.pos[2] = z;
 
   // notify all attached geoms that this body has moved
   for (dxGeom *geom = b->geom; geom; geom = dGeomGetBodyNext (geom))
@@ -325,7 +349,7 @@ void dBodySetRotation (dBodyID b, const dMatrix3 R)
   b->q[1] = q[1];
   b->q[2] = q[2];
   b->q[3] = q[3];
-  dQtoR (b->q,b->R);
+  dQtoR (b->q,b->posr.R);
 
   // notify all attached geoms that this body has moved
   for (dxGeom *geom = b->geom; geom; geom = dGeomGetBodyNext (geom))
@@ -341,7 +365,7 @@ void dBodySetQuaternion (dBodyID b, const dQuaternion q)
   b->q[2] = q[2];
   b->q[3] = q[3];
   dNormalize4 (b->q);
-  dQtoR (b->q,b->R);
+  dQtoR (b->q,b->posr.R);
 
   // notify all attached geoms that this body has moved
   for (dxGeom *geom = b->geom; geom; geom = dGeomGetBodyNext (geom))
@@ -370,14 +394,43 @@ void dBodySetAngularVel (dBodyID b, dReal x, dReal y, dReal z)
 const dReal * dBodyGetPosition (dBodyID b)
 {
   dAASSERT (b);
-  return b->pos;
+  return b->posr.pos;
+}
+
+
+void dBodyCopyPosition (dBodyID b, dVector3 pos)
+{
+	dAASSERT (b);
+	dReal* src = b->posr.pos;
+	pos[0] = src[0];
+	pos[1] = src[1];
+	pos[2] = src[2];
 }
 
 
 const dReal * dBodyGetRotation (dBodyID b)
 {
   dAASSERT (b);
-  return b->R;
+  return b->posr.R;
+}
+
+
+void dBodyCopyRotation (dBodyID b, dMatrix3 R)
+{
+	dAASSERT (b);
+	const dReal* src = b->posr.R;
+	R[0] = src[0];
+	R[1] = src[1];
+	R[2] = src[2];
+	R[3] = src[3];
+	R[4] = src[4];
+	R[5] = src[5];
+	R[6] = src[6];
+	R[7] = src[7];
+	R[8] = src[8];
+	R[9] = src[9];
+	R[10] = src[10];
+	R[11] = src[11];
 }
 
 
@@ -388,6 +441,17 @@ const dReal * dBodyGetQuaternion (dBodyID b)
 }
 
 
+void dBodyCopyQuaternion (dBodyID b, dQuaternion quat)
+{
+	dAASSERT (b);
+	dReal* src = b->q;
+	quat[0] = src[0];
+	quat[1] = src[1];
+	quat[2] = src[2];
+	quat[3] = src[3];
+}
+
+
 const dReal * dBodyGetLinearVel (dBodyID b)
 {
   dAASSERT (b);
@@ -404,10 +468,18 @@ const dReal * dBodyGetAngularVel (dBodyID b)
 
 void dBodySetMass (dBodyID b, const dMass *mass)
 {
-  dAASSERT (b && mass);
+  dAASSERT (b && mass );
+  dIASSERT(dMassCheck(mass));
+
+  // The centre of mass must be at the origin.
+  // Use dMassTranslate( mass, -mass->c[0], -mass->c[1], -mass->c[2] ) to correct it.
+  dUASSERT( fabs( mass->c[0] ) <= dEpsilon &&
+			fabs( mass->c[1] ) <= dEpsilon &&
+			fabs( mass->c[2] ) <= dEpsilon, "The centre of mass must be at the origin." )
+
   memcpy (&b->mass,mass,sizeof(dMass));
   if (dInvertPDMatrix (b->mass.I,b->invI,3)==0) {
-    dDEBUGMSG ("inertia must be positive definite");
+    dDEBUGMSG ("inertia must be positive definite!");
     dRSetIdentity (b->invI);
   }
   b->invMass = dRecip(b->mass.mass);
@@ -447,7 +519,7 @@ void dBodyAddRelForce (dBodyID b, dReal fx, dReal fy, dReal fz)
   t1[1] = fy;
   t1[2] = fz;
   t1[3] = 0;
-  dMULTIPLY0_331 (t2,b->R,t1);
+  dMULTIPLY0_331 (t2,b->posr.R,t1);
   b->facc[0] += t2[0];
   b->facc[1] += t2[1];
   b->facc[2] += t2[2];
@@ -462,7 +534,7 @@ void dBodyAddRelTorque (dBodyID b, dReal fx, dReal fy, dReal fz)
   t1[1] = fy;
   t1[2] = fz;
   t1[3] = 0;
-  dMULTIPLY0_331 (t2,b->R,t1);
+  dMULTIPLY0_331 (t2,b->posr.R,t1);
   b->tacc[0] += t2[0];
   b->tacc[1] += t2[1];
   b->tacc[2] += t2[2];
@@ -480,9 +552,9 @@ void dBodyAddForceAtPos (dBodyID b, dReal fx, dReal fy, dReal fz,
   f[0] = fx;
   f[1] = fy;
   f[2] = fz;
-  q[0] = px - b->pos[0];
-  q[1] = py - b->pos[1];
-  q[2] = pz - b->pos[2];
+  q[0] = px - b->posr.pos[0];
+  q[1] = py - b->posr.pos[1];
+  q[2] = pz - b->posr.pos[2];
   dCROSS (b->tacc,+=,q,f);
 }
 
@@ -500,7 +572,7 @@ void dBodyAddForceAtRelPos (dBodyID b, dReal fx, dReal fy, dReal fz,
   prel[1] = py;
   prel[2] = pz;
   prel[3] = 0;
-  dMULTIPLY0_331 (p,b->R,prel);
+  dMULTIPLY0_331 (p,b->posr.R,prel);
   b->facc[0] += f[0];
   b->facc[1] += f[1];
   b->facc[2] += f[2];
@@ -517,14 +589,14 @@ void dBodyAddRelForceAtPos (dBodyID b, dReal fx, dReal fy, dReal fz,
   frel[1] = fy;
   frel[2] = fz;
   frel[3] = 0;
-  dMULTIPLY0_331 (f,b->R,frel);
+  dMULTIPLY0_331 (f,b->posr.R,frel);
   b->facc[0] += f[0];
   b->facc[1] += f[1];
   b->facc[2] += f[2];
   dVector3 q;
-  q[0] = px - b->pos[0];
-  q[1] = py - b->pos[1];
-  q[2] = pz - b->pos[2];
+  q[0] = px - b->posr.pos[0];
+  q[1] = py - b->posr.pos[1];
+  q[2] = pz - b->posr.pos[2];
   dCROSS (b->tacc,+=,q,f);
 }
 
@@ -542,8 +614,8 @@ void dBodyAddRelForceAtRelPos (dBodyID b, dReal fx, dReal fy, dReal fz,
   prel[1] = py;
   prel[2] = pz;
   prel[3] = 0;
-  dMULTIPLY0_331 (f,b->R,frel);
-  dMULTIPLY0_331 (p,b->R,prel);
+  dMULTIPLY0_331 (f,b->posr.R,frel);
+  dMULTIPLY0_331 (p,b->posr.R,prel);
   b->facc[0] += f[0];
   b->facc[1] += f[1];
   b->facc[2] += f[2];
@@ -592,10 +664,10 @@ void dBodyGetRelPointPos (dBodyID b, dReal px, dReal py, dReal pz,
   prel[1] = py;
   prel[2] = pz;
   prel[3] = 0;
-  dMULTIPLY0_331 (p,b->R,prel);
-  result[0] = p[0] + b->pos[0];
-  result[1] = p[1] + b->pos[1];
-  result[2] = p[2] + b->pos[2];
+  dMULTIPLY0_331 (p,b->posr.R,prel);
+  result[0] = p[0] + b->posr.pos[0];
+  result[1] = p[1] + b->posr.pos[1];
+  result[2] = p[2] + b->posr.pos[2];
 }
 
 
@@ -608,7 +680,7 @@ void dBodyGetRelPointVel (dBodyID b, dReal px, dReal py, dReal pz,
   prel[1] = py;
   prel[2] = pz;
   prel[3] = 0;
-  dMULTIPLY0_331 (p,b->R,prel);
+  dMULTIPLY0_331 (p,b->posr.R,prel);
   result[0] = b->lvel[0];
   result[1] = b->lvel[1];
   result[2] = b->lvel[2];
@@ -621,9 +693,9 @@ void dBodyGetPointVel (dBodyID b, dReal px, dReal py, dReal pz,
 {
   dAASSERT (b);
   dVector3 p;
-  p[0] = px - b->pos[0];
-  p[1] = py - b->pos[1];
-  p[2] = pz - b->pos[2];
+  p[0] = px - b->posr.pos[0];
+  p[1] = py - b->posr.pos[1];
+  p[2] = pz - b->posr.pos[2];
   p[3] = 0;
   result[0] = b->lvel[0];
   result[1] = b->lvel[1];
@@ -637,11 +709,11 @@ void dBodyGetPosRelPoint (dBodyID b, dReal px, dReal py, dReal pz,
 {
   dAASSERT (b);
   dVector3 prel;
-  prel[0] = px - b->pos[0];
-  prel[1] = py - b->pos[1];
-  prel[2] = pz - b->pos[2];
+  prel[0] = px - b->posr.pos[0];
+  prel[1] = py - b->posr.pos[1];
+  prel[2] = pz - b->posr.pos[2];
   prel[3] = 0;
-  dMULTIPLY1_331 (result,b->R,prel);
+  dMULTIPLY1_331 (result,b->posr.R,prel);
 }
 
 
@@ -654,7 +726,7 @@ void dBodyVectorToWorld (dBodyID b, dReal px, dReal py, dReal pz,
   p[1] = py;
   p[2] = pz;
   p[3] = 0;
-  dMULTIPLY0_331 (result,b->R,p);
+  dMULTIPLY0_331 (result,b->posr.R,p);
 }
 
 
@@ -667,7 +739,7 @@ void dBodyVectorFromWorld (dBodyID b, dReal px, dReal py, dReal pz,
   p[1] = py;
   p[2] = pz;
   p[3] = 0;
-  dMULTIPLY1_331 (result,b->R,p);
+  dMULTIPLY1_331 (result,b->posr.R,p);
 }
 
 
@@ -743,6 +815,7 @@ void dBodyEnable (dBodyID b)
   b->flags &= ~dxBodyDisabled;
   b->adis_stepsleft = b->adis.idle_steps;
   b->adis_timeleft = b->adis.idle_time;
+  // no code for average-processing needed here
 }
 
 
@@ -780,28 +853,66 @@ int dBodyGetGravityMode (dBodyID b)
 dReal dBodyGetAutoDisableLinearThreshold (dBodyID b)
 {
 	dAASSERT(b);
-	return dSqrt (b->adis.linear_threshold);
+	return dSqrt (b->adis.linear_average_threshold);
 }
 
 
-void dBodySetAutoDisableLinearThreshold (dBodyID b, dReal linear_threshold)
+void dBodySetAutoDisableLinearThreshold (dBodyID b, dReal linear_average_threshold)
 {
 	dAASSERT(b);
-	b->adis.linear_threshold = linear_threshold * linear_threshold;
+	b->adis.linear_average_threshold = linear_average_threshold * linear_average_threshold;
 }
 
 
 dReal dBodyGetAutoDisableAngularThreshold (dBodyID b)
 {
 	dAASSERT(b);
-	return dSqrt (b->adis.angular_threshold);
+	return dSqrt (b->adis.angular_average_threshold);
 }
 
 
-void dBodySetAutoDisableAngularThreshold (dBodyID b, dReal angular_threshold)
+void dBodySetAutoDisableAngularThreshold (dBodyID b, dReal angular_average_threshold)
 {
 	dAASSERT(b);
-	b->adis.angular_threshold = angular_threshold * angular_threshold;
+	b->adis.angular_average_threshold = angular_average_threshold * angular_average_threshold;
+}
+
+
+int dBodyGetAutoDisableAverageSamplesCount (dBodyID b)
+{
+	dAASSERT(b);
+	return b->adis.average_samples;
+}
+
+
+void dBodySetAutoDisableAverageSamplesCount (dBodyID b, unsigned int average_samples_count)
+{
+	dAASSERT(b);
+	b->adis.average_samples = average_samples_count;
+	// update the average buffers
+	if(b->average_lvel_buffer)
+	{
+		delete[] b->average_lvel_buffer;
+		b->average_lvel_buffer = 0;
+	}
+	if(b->average_avel_buffer)
+	{
+		delete[] b->average_avel_buffer;
+		b->average_avel_buffer = 0;
+	}
+	if(b->adis.average_samples > 0)
+	{
+		b->average_lvel_buffer = new dVector3[b->adis.average_samples];
+		b->average_avel_buffer = new dVector3[b->adis.average_samples];
+	}
+	else
+	{
+		b->average_lvel_buffer = 0;
+		b->average_avel_buffer = 0;
+	}
+	// new buffer is empty
+	b->average_counter = 0;
+	b->average_ready = 0;
 }
 
 
@@ -843,8 +954,20 @@ int dBodyGetAutoDisableFlag (dBodyID b)
 void dBodySetAutoDisableFlag (dBodyID b, int do_auto_disable)
 {
 	dAASSERT(b);
-	if (!do_auto_disable) b->flags &= ~dxBodyAutoDisable;
-	else b->flags |= dxBodyAutoDisable;
+	if (!do_auto_disable)
+	{
+		b->flags &= ~dxBodyAutoDisable;
+		// (mg) we should also reset the IsDisabled state to correspond to the DoDisabling flag
+		b->flags &= ~dxBodyDisabled;
+		b->adis.idle_steps = dWorldGetAutoDisableSteps(b->world);
+		b->adis.idle_time = dWorldGetAutoDisableTime(b->world);
+		// resetting the average calculations too
+		dBodySetAutoDisableAverageSamplesCount(b, dWorldGetAutoDisableAverageSamplesCount(b->world) );
+	}
+	else
+	{
+		b->flags |= dxBodyAutoDisable;
+	}
 }
 
 
@@ -942,6 +1065,11 @@ dxJoint * dJointCreateUniversal (dWorldID w, dJointGroupID group)
   return createJoint (w,group,&__duniversal_vtable);
 }
 
+dxJoint * dJointCreatePR (dWorldID w, dJointGroupID group)
+{
+  dAASSERT (w);
+  return createJoint (w,group,&__dPR_vtable);
+}
 
 dxJoint * dJointCreateFixed (dWorldID w, dJointGroupID group)
 {
@@ -963,6 +1091,17 @@ dxJoint * dJointCreateAMotor (dWorldID w, dJointGroupID group)
   return createJoint (w,group,&__damotor_vtable);
 }
 
+dxJoint * dJointCreateLMotor (dWorldID w, dJointGroupID group)
+{
+  dAASSERT (w);
+  return createJoint (w,group,&__dlmotor_vtable);
+}
+
+dxJoint * dJointCreatePlane2D (dWorldID w, dJointGroupID group)
+{
+  dAASSERT (w);
+  return createJoint (w,group,&__dplane2d_vtable);
+}
 
 void dJointDestroy (dxJoint *j)
 {
@@ -1115,6 +1254,59 @@ dJointFeedback *dJointGetFeedback (dxJoint *joint)
 }
 
 
+
+dJointID dConnectingJoint (dBodyID in_b1, dBodyID in_b2)
+{
+    dAASSERT (in_b1 || in_b2);
+
+	dBodyID b1, b2;
+
+	if (in_b1 == 0) {
+		b1 = in_b2;
+		b2 = in_b1;
+	}
+	else {
+		b1 = in_b1;
+		b2 = in_b2;
+	}
+
+    // look through b1's neighbour list for b2
+    for (dxJointNode *n=b1->firstjoint; n; n=n->next) {
+        if (n->body == b2) return n->joint;
+    }
+
+    return 0;
+}
+
+
+
+int dConnectingJointList (dBodyID in_b1, dBodyID in_b2, dJointID* out_list)
+{
+    dAASSERT (in_b1 || in_b2);
+
+
+	dBodyID b1, b2;
+
+	if (in_b1 == 0) {
+		b1 = in_b2;
+		b2 = in_b1;
+	}
+	else {
+		b1 = in_b1;
+		b2 = in_b2;
+	}
+
+    // look through b1's neighbour list for b2
+    int numConnectingJoints = 0;
+    for (dxJointNode *n=b1->firstjoint; n; n=n->next) {
+        if (n->body == b2)
+            out_list[numConnectingJoints++] = n->joint;
+    }
+
+    return numConnectingJoints;
+}
+
+
 int dAreConnected (dBodyID b1, dBodyID b2)
 {
   dAASSERT (b1 && b2);
@@ -1156,11 +1348,12 @@ dxWorld * dWorldCreate()
   #error dSINGLE or dDOUBLE must be defined
 #endif
 
-  w->adis.linear_threshold = REAL(0.001)*REAL(0.001);	// (magnitude squared)
-  w->adis.angular_threshold = REAL(0.001)*REAL(0.001);	// (magnitude squared)
   w->adis.idle_steps = 10;
   w->adis.idle_time = 0;
   w->adis_flag = 0;
+  w->adis.average_samples = 1;		// Default is 1 sample => Instantaneous velocity
+  w->adis.angular_average_threshold = REAL(0.01)*REAL(0.01);	// (magnitude squared)
+  w->adis.linear_average_threshold = REAL(0.01)*REAL(0.01);		// (magnitude squared)
 
   w->qs.num_iterations = 20;
   w->qs.w = REAL(1.3);
@@ -1179,7 +1372,17 @@ void dWorldDestroy (dxWorld *w)
   dxBody *nextb, *b = w->firstbody;
   while (b) {
     nextb = (dxBody*) b->next;
-    delete b;
+    if(b->average_lvel_buffer)
+    {
+      delete[] (b->average_lvel_buffer);
+      b->average_lvel_buffer = 0;
+    }
+    if(b->average_avel_buffer)
+    {
+      delete[] (b->average_avel_buffer);
+      b->average_avel_buffer = 0;
+    }
+    dBodyDestroy(b); // calling here dBodyDestroy for correct destroying! (i.e. the average buffers)
     b = nextb;
   }
   dxJoint *nextj, *j = w->firstjoint;
@@ -1283,28 +1486,42 @@ void dWorldImpulseToForce (dWorldID w, dReal stepsize,
 dReal dWorldGetAutoDisableLinearThreshold (dWorldID w)
 {
 	dAASSERT(w);
-	return dSqrt (w->adis.linear_threshold);
+	return dSqrt (w->adis.linear_average_threshold);
 }
 
 
-void dWorldSetAutoDisableLinearThreshold (dWorldID w, dReal linear_threshold)
+void dWorldSetAutoDisableLinearThreshold (dWorldID w, dReal linear_average_threshold)
 {
 	dAASSERT(w);
-	w->adis.linear_threshold = linear_threshold * linear_threshold;
+	w->adis.linear_average_threshold = linear_average_threshold * linear_average_threshold;
 }
 
 
 dReal dWorldGetAutoDisableAngularThreshold (dWorldID w)
 {
 	dAASSERT(w);
-	return dSqrt (w->adis.angular_threshold);
+	return dSqrt (w->adis.angular_average_threshold);
 }
 
 
-void dWorldSetAutoDisableAngularThreshold (dWorldID w, dReal angular_threshold)
+void dWorldSetAutoDisableAngularThreshold (dWorldID w, dReal angular_average_threshold)
 {
 	dAASSERT(w);
-	w->adis.angular_threshold = angular_threshold * angular_threshold;
+	w->adis.angular_average_threshold = angular_average_threshold * angular_average_threshold;
+}
+
+
+int dWorldGetAutoDisableAverageSamplesCount (dWorldID w)
+{
+	dAASSERT(w);
+	return w->adis.average_samples;
+}
+
+
+void dWorldSetAutoDisableAverageSamplesCount (dWorldID w, unsigned int average_samples_count)
+{
+	dAASSERT(w);
+	w->adis.average_samples = average_samples_count;
 }
 
 
